self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "313e35c41653be0a749734ba5ee688e3",
    "url": "./index.html"
  },
  {
    "revision": "6b8f413ad69b315a332c",
    "url": "./static/css/2.103a785d.chunk.css"
  },
  {
    "revision": "b8f5c26014c53860f9d7",
    "url": "./static/css/main.2320ee1e.chunk.css"
  },
  {
    "revision": "6b8f413ad69b315a332c",
    "url": "./static/js/2.98d5d6e7.chunk.js"
  },
  {
    "revision": "2ee263441ca071ee7c143c04fef4016c",
    "url": "./static/js/2.98d5d6e7.chunk.js.LICENSE.txt"
  },
  {
    "revision": "b8f5c26014c53860f9d7",
    "url": "./static/js/main.9b488bdc.chunk.js"
  },
  {
    "revision": "673909de58f625d5bba8",
    "url": "./static/js/runtime-main.2b0a6f63.js"
  },
  {
    "revision": "8ad8988c1f5e59c971c764c31becc393",
    "url": "./static/media/draggable_icon.8ad8988c.svg"
  },
  {
    "revision": "63cd8448b347d74fdeb83466c1efcfe1",
    "url": "./static/media/icon-downarrow.63cd8448.svg"
  },
  {
    "revision": "823f5cc21758b9a495f031bcc3ab18b1",
    "url": "./static/media/icon-uparrow.823f5cc2.svg"
  },
  {
    "revision": "09f1ebe8acecfb26c3e24a9b1a6de64a",
    "url": "./static/media/icon_success.09f1ebe8.svg"
  },
  {
    "revision": "1c2571e3cace18d5b9813f8251f55893",
    "url": "./static/media/loading_icon.1c2571e3.gif"
  },
  {
    "revision": "99f737800bf8cc41fbb52953f8e89f89",
    "url": "./static/media/sort-indicator-desc-icon.99f73780.svg"
  },
  {
    "revision": "6e7e2adc7f5c2ab7646470717a86835c",
    "url": "./static/media/sort-indicator-none-icon.6e7e2adc.svg"
  }
]);